<?php

namespace Ems\Services;

use Doctrine\ORM\EntityManager;
use Ems\Repository\{ ApuracaoRepository, PeriodoRepository, MarcacaoRepository, OcorrenciaRepository, VinculoRepository };
use Ems\Services\{ OcorrenciaService, VinculoService };
use Ems\Util\ArrayUtil;
use Ems\Util\Time;

use Util\{ Environment, Messages, DateUtil };
use Folha\Logic\ControlePonto\Apuracao;

class ApuracaoService {
    /** @var ApuracaoRepository $apuracaoRepository */
    protected $apuracaoRepository;

    /** @var Environment $environment */
    protected $environment;

    protected $periodoRepository;

    /** @var EntityManager $entityManager */
    protected $entityManager;

    /** @var VinculoRepository */
    protected $vinculoRepository;
    
    
    // Constants usadas para validação
    const APUR_TEMPO_MAXIMO_SEM_ALMOCO          = 6;
    const APUR_TEMPO_MINIMO_ALMOCO              = 1;
    const APUR_TEMPO_MIN_SEM_ALMOCO             = 4;
    
    const TEMPO_JORNADA_INTERVALO_15_TEMPO_MAX  = 6;
    const TEMPO_JORNADA_INTERVALO_15_TEMPO_MIN  = 4;
    
    const TEMPO_MINIMO_INTERVALO                = 0.25;
    
    const APUR_JORNADA_EST                      = 6;
    const APUR_INTERVALO_JORNADA_EST            = 15 / 60;
    
    // Tipos de vínculo cujo processo de apuração é realizado dia a dia
    const TIPOS_VINC_APUR_DIARIA = [VinculoService::TIPO_VINC_EST];

    public function __construct(
        Environment             $environment,
        ApuracaoRepository      $apuracaoRepository,
        EntityManager           $entityManager,
        MarcacaoPontoService    $marcacaoPontoService,
        VinculoService          $vinculoService,
        PeriodoRepository       $periodoRepository,
        MarcacaoRepository      $marcacaoRepository,
        OcorrenciaRepository    $ocorrenciaRepository,
        VinculoRepository       $vinculoRepository
    ) {

        $this->environment          = $environment;
        $this->apuracaoRepository   = $apuracaoRepository;
        $this->entityManager        = $entityManager;
        $this->marcacaoPontoService = $marcacaoPontoService;
        $this->vinculoService       = $vinculoService;
        $this->ocorrenciaRepository = $ocorrenciaRepository;
        $this->periodoRepository    = $periodoRepository;
        $this->marcacaoRepository   = $marcacaoRepository;
        $this->vinculoRepository    = $vinculoRepository;
    }

    public function retrocederApuracao(
        $nrVinculo,
        $nrPeriodoApuracao
    ) : int {

        $nrOrg = $this->environment->getNrOrg();

        $intervaloPeriodo = $this->periodoRepository->
            getPeriodoIntervalo($nrOrg, $nrPeriodoApuracao);

        $iniPeriodo = $intervaloPeriodo[0]['DTINICIALAPURACAO'];

        $this->apuracaoRepository->deleteMarcacoesPeriodo($nrOrg, $nrVinculo, $nrPeriodoApuracao);
        $this->apuracaoRepository->deleteIntervalosPeriodo($nrOrg, $nrVinculo, $nrPeriodoApuracao);
        $this->apuracaoRepository->deleteResumosPeriodo($nrOrg, $nrVinculo, $nrPeriodoApuracao);

        $this->apuracaoRepository->updateMarcacoesAlteradasPeriodo($nrOrg, $nrVinculo, $nrPeriodoApuracao);

        $this->apuracaoRepository->updateNovoCodigoResumos($nrOrg);
        $this->apuracaoRepository->updateNovoCodigoMarcacoes($nrOrg);
        $this->apuracaoRepository->updateNovoCodigoIntervalos($nrOrg);

        return 1;
    }
    
    public function retrocederDia(
        $nrVinculo,
        $dia
    ) {
        $nrOrg = $this->environment->getNrOrg();
        
        $this->apuracaoRepository->deleteMarcacoesDia($nrOrg, $nrVinculo, $dia);
        $this->apuracaoRepository->deleteIntervalosDia($nrOrg, $nrVinculo, $dia);
        $this->apuracaoRepository->deleteResumosDia($nrOrg, $nrVinculo, $dia);

        $this->apuracaoRepository->updateMarcacoesAlteradasDia($nrOrg, $nrVinculo, $dia);

        $this->apuracaoRepository->updateNovoCodigoResumos($nrOrg);
        $this->apuracaoRepository->updateNovoCodigoMarcacoes($nrOrg);
        $this->apuracaoRepository->updateNovoCodigoIntervalos($nrOrg);
    }
    
    /**
     *  Rotina responsável por realizar a apuração de um período para um determinado vínculo
     * 
     *  @param  int     $nrVinculo
     *  @param  int     $nrPeriodoApuracao
     *  @param  bool    $ateDiaAtual        = false
     */
    public function apurarPeriodo(
        int  $nrVinculo,
        int  $nrPeriodoApuracao,
        bool $ateDiaAtual = false
    ) {
        $nrOrg = (int) $this->environment->getNrOrg();
        $apurado = false;
        
        if (in_array($this->vinculoService->getTipoVinculo($nrVinculo, $nrOrg), self::TIPOS_VINC_APUR_DIARIA)) {
            $apurado = $this->apurarPeriodoDiaADia(
                $nrVinculo,
                $nrPeriodoApuracao,
                $ateDiaAtual
            );
        } else {
            $apurado = $this->apurarPeriodoCompleto(
                $nrVinculo,
                $nrPeriodoApuracao,
                $ateDiaAtual
            );
        }
        
        return $apurado;
    }

    /**
     *  Rotina que executa o processo de apuração para um período por inteiro
     * 
     *  @param  int     $nrVinculo
     *  @param  int     $nrPeriodoApuracao
     *  @param  bool    $ateDiaAtual        = false
     * 
     *  @return bool    Status
     */
    public function apurarPeriodoCompleto(
        int  $nrVinculo,
        int  $nrPeriodoApuracao,
        bool $ateDiaAtual = false
    ) {
        $nrOrg = (int) $this->environment->getNrOrg();

        $this->marcacaoPontoService->validaTempoMarcacoesPeriodo($nrVinculo, $nrOrg, $nrPeriodoApuracao);

        $intervaloPeriodo = $this->periodoRepository->
            getPeriodoIntervalo($nrOrg, $nrPeriodoApuracao);

        $iniPeriodo = $intervaloPeriodo[0]['DTINICIALAPURACAO'];
        $fimPeriodo = $intervaloPeriodo[0]['DTFINALAPURACAO'];

        $this->retrocederApuracao($nrVinculo, $nrPeriodoApuracao);

        if ($ateDiaAtual) {
            $periodoAtual = $this->getPeriodoAtual($nrVinculo);
            $nrPeriodoAtual = (int) $periodoAtual['NRPERIODOAPURACAO'];

            if ($nrPeriodoAtual === $nrPeriodoApuracao) {
                $dataAtual = DateUtil::getDataAtual()->format('d/m/Y');

                $fimPeriodo = $dataAtual;
            }
        }

        $this->realizaApuracao(
            $nrOrg,
            $nrVinculo,
            $iniPeriodo,
            $fimPeriodo,
            $nrPeriodoApuracao
        );

        return true;
    }
    
    /**
     *  Rotina criada para contemplar casos em que é necessário realizar a apuração dia a dia, como o caso do estagiário
     * 
     *  @param  int     $nrVinculo
     *  @param  int     $nrPeriodoApuracao
     *  @param  bool    $ateDiaAtual        = false
     * 
     *  @return bool    Status
     */
    public function apurarPeriodoDiaADia(
        int  $nrVinculo,
        int  $nrPeriodoApuracao,
        bool $ateDiaAtual = false
    ) {
        $nrOrg = (int) $this->environment->getNrOrg();
        
        $periodo = $this->periodoRepository->getPeriodo($nrOrg, $nrPeriodoApuracao);
        if (empty($periodo)) {
            throw new \Exception(Messages::PERIODO_PER_NOT_FOUND);
        }
        
        // Retrocede se houver resumos
        $temResumo = $this->verificaPeriodoTemResumo($nrVinculo, $nrPeriodoApuracao);

        if ($temResumo) {
            $this->retrocederApuracao($nrVinculo, $nrPeriodoApuracao);
        }

        // Tranforma datas para datetimes
        $dataInicialPeriodo = DateUtil::getDataDeString($periodo['DTINICIALAPURACAO'], DateUtil::FORMATO_DATAHORA, true);
        $dataFinalPeriodo = DateUtil::getDataDeString($periodo['DTFINALAPURACAO'], DateUtil::FORMATO_DATAHORA, true);

        // Para cada dia do período, invoca a validação da rotina self::validaTempoMarcacoes
        $dataCorrente = $dataInicialPeriodo;
        while($dataCorrente <= $dataFinalPeriodo) {
            $this->apurarDia($nrVinculo, $dataCorrente->format(DateUtil::FORMATO_DATA));
            $dataCorrente = DateUtil::adicionaIntervalo($dataCorrente, 1, DateUtil::DIA);
        }

        
        return true;
    }
    
    /**
     * Aciona a apuracao para dias de colaboradores com ponto ativo a partir do inicio do perido ate a data atual
     * 
     * @param string $dataFim
     * @param array $nrvinculomList
     * @param array $nrestrutlegalList
     * @param array $nrestrutgerenList
     * @param array $nrtipovinculomLis
     */
    public function apurarPeriodoAteDataVinculosPontoAtivo(
        ?string $dataFim = null,
        ?array $nrvinculomList = [],
        ?array $nrestrutlegalList = [],
        ?array $nrestrutgerenList = [],
        ?array $nrtipovinculomList = []
    ): void {
        $paramDataFim = $dataFim ?? DateUtil::getDataAtual()->format(DateUtil::FORMATO_DATA);
        $vinculos = $this->apuracaoRepository->getVinculosPontoAtivoAteData(
            $paramDataFim,
            $nrvinculomList,
            $nrestrutlegalList,
            $nrestrutgerenList,
            $nrtipovinculomList
        );
        
        foreach ($vinculos as $vinculo) {
            $datas = explode(',', $vinculo['DATAS_APURAR_PERIODO_ATUAL'] ?? '');
            if (empty($datas)) {
                continue;
            }
            $this->apurarDias($vinculo['NRVINCULOM'], $datas);
        }
    }

    public function getPeriodoAtual(
        int $nrVinculo
    ): array {
        $nrOrg = $this->environment->getNrOrg();

        $vinculo = $this->vinculoRepository->getVinculoHistorico(
            $nrOrg, $nrVinculo
        );

        $nrTipoVinculo = (int) $vinculo['NRTIPOVINCULOM'];

        $periodoParams = [
            'P_NRORG'           => $nrOrg,
            'P_NRTIPOVINCULOM'  => $nrTipoVinculo,
        ];

        $periodoAtual = $this->periodoRepository->getPeriodoAtual(
            $periodoParams
        );

        return $periodoAtual;
    }

    public function apurarDia(
        $nrVinculo,
        $dia
    ) {
        $nrOrg = $this->environment->getNrOrg();
        
        $iniPeriodo = $dia;
        $fimPeriodo = $dia;
        
        $periodo = $this->periodoRepository->findPeriodoByDia(
            $nrOrg,
            $nrVinculo,
            $dia
        );
        
        if (empty($periodo)) {
            throw new \Exception(Messages::PERIODO_PER_NOT_FOUND);
        }

        $this->validacaoProcessoDependeDiasApurados(
            $nrVinculo, [$dia], false, true, false
        );
        
        $nrPeriodoApuracao = (int) $periodo['NRPERIODOAPURACAO'];
        
        // Validação do tempo total das marcações
        $this->marcacaoPontoService->validaTempoMarcacoes(
            $nrVinculo,
            $nrOrg,
            $dia
        );
        
        $this->retrocederDia(
            $nrVinculo,
            $dia
        );
        
        $this->realizaApuracao(
            $nrOrg,
            $nrVinculo,
            $iniPeriodo,
            $fimPeriodo,
            $nrPeriodoApuracao
        );

        return true;
    }
    
    public function apurarDias(
        int     $nrVinculo,
        array   $dias
    ) {
        $dias = DateUtil::sortDateArray($dias);

        foreach ($dias as $dia) {
            $this->apurarDia(
                $nrVinculo,
                $dia
            );
        }
        
        return true;
    }
    
    public function realizaApuracao(
        $nrOrg,
        $nrVinculo,
        $iniPeriodo,
        $fimPeriodo,
        $nrPeriodoApuracao
    ) {
        $apuracao = new Apuracao(
            $nrOrg,
            $nrVinculo,
            $iniPeriodo,
            $fimPeriodo,
            $nrPeriodoApuracao
        );

        return $apuracao->gravaApuracao();    
    }

    public function verificaPeriodoTemResumo(
        int $nrVinculo,
        int $nrPeriodoApuracao
    ) : bool {

        $resumos = $this->getResumoSituFreqVinculo($nrVinculo, $nrPeriodoApuracao);
        $temResumo = !empty($resumos);

        return $temResumo;
    }
    
    public function verificaDiaTemResumo(
        int     $nrVinculo,
        string  $dia
    ) : bool {

        $resumos = $this->getResumoDiaSituFreqVinculo($nrVinculo, $dia);
        $temResumo = !empty($resumos);

        return $temResumo;
    }

    public function getResumoSituFreqVinculo (
        int $nrVinculo,
        int $nrPeriodoApuracao
    ) : array {

        $nrOrg = $this->environment->getNrOrg();

        $params = [
            'P_NRORG' => $nrOrg,
            'P_NRVINCULOM' => $nrVinculo,
            'P_NRPERIODOAPURACAO' => $nrPeriodoApuracao
        ];

        $result = $this->apuracaoRepository->getResumoSituFreqVinculo($params);

        return $result;
    }
    
    public function getResumoDiaSituFreqVinculo (
        int     $nrVinculo,
        string  $dtapuracaositufreq
    ) : array {

        $nrOrg = $this->environment->getNrOrg();

        $params = [
            'P_NRORG'               => $nrOrg,
            'P_NRVINCULOM'          => $nrVinculo,
            'P_DTAPURACAOSITUFREQ'  => $dtapuracaositufreq
        ];

        $result = $this->apuracaoRepository->getResumoDiaSituFreqVinculo($params);

        return $result;
    }

    public function getAlteEscala(int $nrVinculo, string $dia) {
        $nrOrg = $this->environment->getNrOrg();

        $params = [
            'P_NRORG'      => $nrOrg,
            'P_NRVINCULOM' => $nrVinculo,
            'P_DATE'        => $dia
        ];

        $alteEscala = $this->apuracaoRepository->getAlteEscala($params);

        return isset($alteEscala[0]) ? $alteEscala[0] : false;
    }
    
    /**
     *  Rotina útil para qualquer ação que dependa da APURAÇÃO de dias anteriores aos que foram passados sejam previamente apurados dentro do dos PERIODOS ABERTOS. Retorna os dias ordenados cronológicamente.
     * 
     * @param   Array $dates | Array com os dias como string 'd/m/Y'
     * 
     */
    public function validacaoProcessoDependeDiasApurados(
                $nrVinculo,
        array   $dates,
        bool    $validaSeDiasNaoPossuemMarcacao = true,
        bool    $throwException = true,
        bool    $considerDay = true // verifica se o dia enviado também possui apuração
    ) {
        $status = [
            'success' => true,
            'message' => null
        ];

        $nrorg = $this->environment->getNrOrg();

        // Realiza validação se dias possuem marcações. Se tiverem, uma exceção é levantada
        if ($validaSeDiasNaoPossuemMarcacao) {
            $marcacoesNosDiasSelecionados = $this->marcacaoRepository->getMarcacaoByDiaMarcacao(
                $nrorg,
                $nrVinculo,
                $dates
            );

            $existemMarcacoesNosDiasSelecionados = !empty($marcacoesNosDiasSelecionados);

            if ($existemMarcacoesNosDiasSelecionados) {
                $message = Messages::APUR_PROCESSO_NAO_REALIZADO_MARCACOES_EXISTENTES;

                if ($throwException) {
                    throw new \Exception($message);
                } else {
                    $status['success'] = false;
                    $status['message'] = $message;

                    return $status;
                }
            }
        }

        // Realiza validação se dias anteriores foram apurados. Se houverem dias com marcação que não foram apurados, uma exceção é levantada.
        $ultPeriodoAberto = $this->periodoRepository->getDataInicialUltPeriodoAberto(
            $nrorg,
            $nrVinculo
        );

        $primeiroDiaUltPeriodoAberto = $ultPeriodoAberto['DTINICIALAPURACAO'];

        $primeiroDia = $primeiroDiaUltPeriodoAberto;

        $ultimoDia = DateUtil::getMinDayFromDateArray($dates)->format('d/m/Y');

        // Dias a partir do primeiro dia do periodo aberto mais antigo até o último selecionado
        $existemDiasNaoApurados = $this->existemDiasNaoApurados(
            $nrorg,
            $nrVinculo,
            $primeiroDia,
            $ultimoDia,
            $considerDay
        );

        if ($existemDiasNaoApurados) {
            $message = Messages::APUR_PROCESSO_NAO_REALIZADO_PENDENCIAS_APURACAO;

            if ($throwException) {
                throw new \Exception($message);
            } else {
                $status['success'] = false;
                $status['message'] = $message;

                return $status;
            }
        }

        return $status;
    }
    
    public function existemDiasNaoApurados(
                $nrOrg,
                $nrVinculo,
        string  $primeiroDia,
        string  $ultimoDia = null,
        bool    $considerDay = true
    ): bool {

        $diasNaoApurados = $this->apuracaoRepository->buscaDiasComMarcacaoNaoApurados(
            $nrOrg,
            $nrVinculo,
            $primeiroDia,
            $ultimoDia
        );

        $dias = array_map(function($diaNaoApurado) {
            return $diaNaoApurado['DTHRMARCACAOPONTO'];
        }, $diasNaoApurados);

        if (!$considerDay && in_array($ultimoDia, $dias)) {
            $index = array_search($ultimoDia, $dias);
            unset($dias[$index]);

            array_values($dias);
        }
        
        $existemDiasNaoApurados = !empty($dias);
        
        return $existemDiasNaoApurados;
    }

    public function getApuracaoPorData(
        int    $nrVinculo,
        string $date
    ): array {
        $nrOrg = (int) $this->environment->getNrOrg();

        $apuracao = $this->apuracaoRepository->getApuracaoPorData(
            $nrVinculo, $nrOrg, $date
        );

        return $apuracao;
    }

    public function apurarPeriodoFuncionario(
        int $nrVinculo,
        int $nrPeriodo
    ) {
        $this->apurarPeriodo($nrVinculo, $nrPeriodo, true);
    }
    
    /**
     *  Rotina responsável por retornar a diferença apurada de tempo para um dia especifico
     * 
     *  @param  int    $nrVinculo
     *  @param  string $date
     * 
     *  @return Time    $diferencaDia
     * 
     */
    public function getDiferencaDia(
        int     $nrVinculo,
        string  $date
    ) {
        
        $apuracao = $this->getApuracaoPorData(
            $nrVinculo, $date
        );
        
        if (empty($apuracao)) {
            return false;
        }
        
        $diferencaDia = floatval($apuracao['DIFERENCAJORNADA']) * 60;
        
        return new Time($diferencaDia);
    }
    
    /**
     *  Rotina responsável por reapurar dias que já possuem apuração, após um dia específico, dentro do respectivo período em que o dia se encontra
     *
     *  @param          $nrVinculo
     *  @param  string  $dia
     *
     *  @return bool    success
     */
    public function atualizaApuracaoDiasPeriodo (
                $nrVinculo,
        string  $dia
    ) {
        $diasApuradosPosterioresPeriodo = $this->getDiasPosterioresNaoApurados($nrVinculo, $dia);

        if (empty($diasApuradosPosterioresPeriodo)) {
            return true;
        }

        return $this->apurarDias($nrVinculo, $diasApuradosPosterioresPeriodo);
    }

    /**
     *  Rotina responsável por obter os dias que já possuem apuração, após um dia específico, dentro do respectivo período em que o dia se encontra
     *
     *  @param          $nrVinculo
     *  @param  string  $dia
     *
     *  @return Array   ApuracaoRepository\getDiasPosterioresNaoApurados
     */
    public function getDiasPosterioresNaoApurados (
                $nrVinculo,
        string  $data
    ) {
        $nrOrg = $this->environment->getNrorgTrab();
        $diasQueryColumn = 'DTAPURACAOSITUFREQ';

        $dias = $this->apuracaoRepository->getDiasPosterioresNaoApurados($nrOrg, $nrVinculo, $data);

        return array_map( function($dia) use ($diasQueryColumn) {
            return $dia[$diasQueryColumn];
        }, $dias);
    }

    public function apurarDiasPosteriores(
        int $nrVinculo,
        string $date
    ) {
        $dias = $this->getDiasPosterioresNaoApurados(
            $nrVinculo, $date
        );

        $diasPosteriores = array_filter($dias, function($dia) use ($date) {
            return $dia !== $date;
        });

        if (empty($diasPosteriores)) {
            return true;
        }

        return $this->apurarDias($nrVinculo, $diasPosteriores);
    }
}